import json
from pathlib import Path

import pytest
from apischema import deserialize
from pytest_mock import MockerFixture
from requests import HTTPError, Response

from mock_event_generator import GEventFetcher
from mock_event_generator.exceptions import MEGIgnoreEventError
from mock_event_generator.gracedbs import GraceDBAlias, GraceDBWithContext
from mock_event_generator.models import GEventDescription
from pytest_gracedb import PIPELINES, Pipeline

GEVENT_ID = PIPELINES['cwb'].gevent_id
SOURCE = GraceDBWithContext.meg_from_alias_or_url(GraceDBAlias.MOCKED)


@pytest.mark.parametrize('pipeline_id, pipeline', PIPELINES.items())
def test_fetchers(
    mocked_gracedb: None, tmp_path: Path, pipeline_id: str, pipeline: Pipeline
) -> None:
    gevent_id = pipeline.gevent_id
    fetcher = GEventFetcher.from_id(gevent_id, SOURCE, tmp_path)
    fetcher.fetch()

    path = tmp_path / gevent_id
    description_path = path / 'description.json'
    assert description_path.is_file()
    description_serialized = json.loads(description_path.read_text())
    description = deserialize(GEventDescription, description_serialized)
    assert description.id == gevent_id

    expected = {
        'cwb': {
            'trigger_1338848303.7875.txt,0',
            'cWB.fits.gz,0',
        },
        'fermi': {
            'tmpykkr0kg3,0',
            'fermi_test_skymap.fits.gz,0',
            'fermi_test_skymap.fits.gz,1',
            'coincidence_far.json,2',
            'coincidence_far.json,3',
        },
        'gstlal': {
            'coinc.xml,0',
            'p_astro.json,0',
            'ranking_data.xml.gz,0',
        },
        'mbta': {
            'coinc.xml,0',
            'pastro.json,0',
            'embright.json,0',
        },
        'mly': {
            'T_1339447521.0_HLV.json,0',
        },
        'pycbc': {
            'coinc-1339452303.1486003-jnqfly.xml.gz,0',
            'coinc-1339452303.1486003-jnqfly.hdf,0',
            'coinc-1339452303.1486003-jnqfly_probs.json,0',
        },
        'spiir': {
            'H1L1V1_1339427916_398_235.xml,0',
            'psd.xml.gz,0',
        },
    }[pipeline_id]

    assert {_.name for _ in path.glob('*') if '.png,' not in _.name} - {
        'description.json'
    } == expected


@pytest.mark.parametrize('exception_cls', [KeyboardInterrupt, ValueError])
def test_cleanup(
    mocker: MockerFixture,
    mocked_gracedb: None,
    tmp_path: Path,
    exception_cls: type[BaseException],
) -> None:
    mocked_rmtree = mocker.patch('shutil.rmtree')
    mocked_fetch_no_cleanup = mocker.patch(
        'mock_event_generator.GEventFetcher._fetch_no_cleanup',
        autospec=True,
        side_effect=exception_cls,
    )

    with pytest.raises(exception_cls):
        GEventFetcher.from_id(GEVENT_ID, SOURCE, tmp_path).fetch()

    mocked_fetch_no_cleanup.assert_called_once()
    mocked_rmtree.assert_called_once()


def test_cleanup_present(
    mocker: MockerFixture, mocked_gracedb: None, tmp_path: Path
) -> None:
    mocked_rmtree = mocker.patch('shutil.rmtree')
    (tmp_path / GEVENT_ID).mkdir()
    GEventFetcher.from_id(GEVENT_ID, SOURCE, tmp_path).fetch()
    mocked_rmtree.assert_called_once()


def test_cleanup_missing(
    mocker: MockerFixture, mocked_gracedb: None, tmp_path: Path
) -> None:
    mocked_rmtree = mocker.patch('shutil.rmtree')
    GEventFetcher.from_id(GEVENT_ID, SOURCE, tmp_path).fetch()
    mocked_rmtree.assert_not_called()


def test_nologs(mocker: MockerFixture, mocked_gracedb: None, tmp_path: Path) -> None:
    mocked_filter_logs = mocker.patch(
        'mock_event_generator.GEventFetcher._filter_logs', return_value=[]
    )
    fetcher = GEventFetcher.from_id(GEVENT_ID, SOURCE, tmp_path)
    with pytest.raises(MEGIgnoreEventError):
        fetcher.fetch()
    mocked_filter_logs.assert_called_once()


def test_store_file_http_error(
    mocker: MockerFixture, mocked_gracedb: None, tmp_path: Path
) -> None:
    old_raise_for_status = Response.raise_for_status

    def new_raise_for_status(response: Response) -> None:
        assert response.request.url is not None
        if '/files/' in response.request.url:
            response.status_code = 500
        old_raise_for_status(response)

    fetcher = GEventFetcher.from_id(GEVENT_ID, SOURCE, tmp_path)
    mocker.patch(
        'requests.Response.raise_for_status',
        autospec=True,
        side_effect=new_raise_for_status,
    )
    with pytest.raises(HTTPError):
        fetcher.fetch()


def test_store_file_present_gz(
    mocker: MockerFixture, mocked_gracedb: None, tmp_path: Path
) -> None:
    fetcher = GEventFetcher.from_id(GEVENT_ID, SOURCE, tmp_path)
    mocked_store_file = mocker.spy(fetcher, '_store_file')
    filename = PIPELINES['cwb'].files[0]
    fetcher._store_file(tmp_path, filename + '.gz')
    assert mocked_store_file.call_count == 2
    mocked_store_file.assert_called_with(tmp_path, filename)


def test_store_file_missing_gz(
    mocker: MockerFixture, mocked_gracedb: None, tmp_path: Path
) -> None:
    fetcher = GEventFetcher.from_id(GEVENT_ID, SOURCE, tmp_path)
    mocked_store_file = mocker.spy(fetcher, '_store_file')
    with pytest.raises(ValueError):
        fetcher._store_file(tmp_path, 'xxx.gz')
    assert mocked_store_file.call_count == 2
    mocked_store_file.assert_called_with(tmp_path, 'xxx')



import pytest

from mock_event_generator.utils import is_any_event, is_gevent, is_superevent


@pytest.mark.parametrize('id', ['G12345', 'M12345', 'T12345'])
def test_is_gevent(id: str) -> None:
    assert is_gevent(id)


@pytest.mark.parametrize('id', ['S12345a', 'G123', 'G12345a', 'S12345', 'MS12345'])
def test_is_not_gevent(id: str) -> None:
    assert not is_gevent(id)


@pytest.mark.parametrize('id', ['MS12345a', 'S12345a'])
def test_is_superevent(id: str) -> None:
    assert is_superevent(id)


@pytest.mark.parametrize(
    'id', ['G12345', 'S123a', 'MS123a', 'S00000', 'M12345a', 'G12345a']
)
def test_is_not_superevent(id: str) -> None:
    assert not is_superevent(id)


@pytest.mark.parametrize('id', ['G12345', 'M12345', 'T12345', 'MS12345a', 'S12345a'])
def test_is_any_event(id: str) -> None:
    assert is_any_event(id)


@pytest.mark.parametrize(
    'id',
    [
        'G123',
        'G12345a',
        'S12345',
        'MS12345',
        'S123a',
        'MS123a',
        'S00000',
        'M12345a',
        'G12345a',
    ],
)
def test_is_not_anyevent(id: str) -> None:
    assert not is_any_event(id)



"""Classes for creating G-events in GraceDB."""
from __future__ import annotations

import logging
import re
from dataclasses import dataclass
from io import BytesIO
from typing import Any

import trio
from astropy.time import Time
from lxml import etree
from lxml.etree import _Element, _ElementTree
from requests import HTTPError

import ligo.lw
import ligo.lw.array
import ligo.lw.lsctables
import ligo.lw.param
import ligo.lw.table
import ligo.lw.utils
from ligo.lw.ligolw import Document, LIGOLWContentHandler

from ._vendor.ligo.skymap.util import ilwd
from .cache import EventFileCache, GEventCacheEntry, SEventCacheEntry
from .gracedbs import GraceDBWithContext
from .models import GEventDescription, SEventDescription, Upload
from .utils import split_filename

IS_FITS_REGEX = re.compile(r'\.fits(\.gz)?$')

logger = logging.getLogger(__name__)


@ilwd.use_in
@ligo.lw.array.use_in
@ligo.lw.lsctables.use_in
@ligo.lw.param.use_in
@ligo.lw.table.use_in
class ContentHandler(LIGOLWContentHandler):  # type: ignore[misc]
    """Class for parsing LIGO Light Weight documents with a SAX2-compliant parser.

    Old-style row IDs are converted on the fly.
    """


@dataclass
class SEventCreator:
    """Creator of the G-events that belong to a given S-event in GraceDB.

    It creates G-events mostly identitical to the G-events belonging to the original
    S-event. The only G-event fields that can be modified are:
        - the time reference, which is translated to the current time.
        - `group` the analysis group which identified the candidate.
        - `search` the type of search of the analysis pipeline.

    Attributes:
        sevent: The original S-event description.
        target: The GraceDB client where the G-events will be created.
        cache_entry: The cache entry, which stores the original G-event' data files.
    """

    sevent: SEventDescription
    target: GraceDBWithContext
    cache_entry: SEventCacheEntry

    @classmethod
    def from_id(
        cls, sevent_id: str, target: GraceDBWithContext, cache: EventFileCache
    ) -> SEventCreator:
        """Factory to create the G-events that belong to a given S-event.

        Parameters:
            sevent_id: Identifier of the S-event to be downloaded.
            target: The GraceDB client where the G-events will be created.
            cache: The event data file cache.
        """
        cache_entry = cache.get_sevent_cache_entry(sevent_id)
        sevent = cache_entry.get_description()
        return SEventCreator(sevent, target, cache_entry)

    async def create(
        self, group: str | None, search: str | None, max_delay: float | None = None
    ) -> list[dict[str, Any]]:
        """Time-translate and create G-events.

        Time-translate all the G-events belonging to the original S-event and create
        them in GraceDB.

        Parameters:
            group: The analysis group which identified the candidate. If None, the
                original group is used.
            search: The type of search of the analysis pipeline. If None, the
                original search type is used.
            max_delay: If not None, maximum lapse of time between the creation of the
                first G-Event and the last upload. If None, the original time sequencing
                of the uploads is reproduced.
        """
        creators = self._get_creators()
        if not creators:
            logger.warning('No events have been created.')
            return []

        gevents_in = [_.gevent for _ in creators]
        upload_delays = self._get_upload_delays(gevents_in, max_delay)
        min_creation_time = min(self._get_original_creation_times(gevents_in))
        now = Time.now().gps

        async def func(
            creator: GEventCreator, event_time: float, request_delay: float
        ) -> None:
            id = creator.gevent.id
            gevents_out[id] = await creator.create(
                group, search, event_time, request_delay
            )

        gevents_out: dict[str, dict[str, Any]] = {}
        async with trio.open_nursery() as nursery:
            for creator in creators:
                creation_delay = upload_delays[creator.gevent.id][0]
                event_time = now - min_creation_time + creator.gevent.gpstime
                nursery.start_soon(func, creator, event_time, creation_delay)

        return list(gevents_out.values())

    def _get_creators(self) -> list[GEventCreator]:
        creators: list[GEventCreator] = []
        for gevent_id in self.sevent.gevent_ids:
            cache_entry = self.cache_entry / gevent_id
            gevent = cache_entry.get_description()
            creator_cls = PIPELINE_GEVENT_CREATOR_CLASSES.get(gevent.pipeline)
            if creator_cls is None:
                logger.warning(
                    f'Cannot re-create G-event {gevent_id!r} from unknown pipeline '
                    f'{gevent.pipeline!r}.'
                )
                continue
            creator = creator_cls(gevent, self.target, cache_entry)
            creators.append(creator)
        return creators

    def _get_upload_delays(
        self, gevents: list[GEventDescription], max_delay: float | None
    ) -> dict[str, list[float]]:
        original_delays = self._get_original_upload_delays(gevents)
        shrink_factor = self._get_shrink_factor(original_delays, max_delay)
        return {
            id: [_ * shrink_factor for _ in delays]
            for id, delays in original_delays.items()
        }

    def _get_original_upload_delays(
        self, gevents: list[GEventDescription]
    ) -> dict[str, list[float]]:
        creation_times = self._get_original_creation_times(gevents)
        min_creation_time = min(creation_times)

        upload_delays = {
            gevent.id: [
                creation_time - min_creation_time + upload.delay
                for upload in filter_uploads_unless(gevent.uploads, False)
            ]
            for gevent, creation_time in zip(gevents, creation_times)
        }
        return upload_delays

    def _get_original_creation_times(
        self, gevents: list[GEventDescription]
    ) -> list[float]:
        return [_.gpstime + _.reporting_latency for _ in gevents]

    @staticmethod
    def _get_shrink_factor(
        upload_delays: dict[str, list[float]], max_delay: float | None
    ) -> float:
        actual_max_delay = max(delays[-1] for delays in upload_delays.values())
        if max_delay is None:
            logger.info(
                f'The uploads to GraceDB will take {actual_max_delay:.2f}s, like the '
                f'original schedule.'
            )
            return 1.0

        if actual_max_delay == 0:
            return 1.0

        return max_delay / actual_max_delay


@dataclass
class GEventCreator:
    """Creator of G-events in GraceDB.

    It creates G-events mostly identitical to the specified input G-event. The only
    G-event fields that are / can be modified are:
        - the time reference, which is translated to the current time.
        - `group` the analysis group which identified the candidate.
        - `search` the type of search of the analysis pipeline.

    Attributes:
        gevent: The original G-event description.
        target: The GraceDB client where the G-event will be created.
        cache_entry: The cache entry, which stores the original G-event data files.
    """

    gevent: GEventDescription
    target: GraceDBWithContext
    cache_entry: GEventCacheEntry

    @classmethod
    def from_id(
        cls, gevent_id: str, target: GraceDBWithContext, cache: EventFileCache
    ) -> GEventCreator:
        """Factory to create G-events.

        Parameters:
            gevent_id: Identifier of the G-event to be downloaded.
            target: The GraceDB client where the G-event will be created.
            cache: The event data file cache.
        """
        cache_entry = cache.get_gevent_cache_entry(gevent_id)
        gevent = cache_entry.get_description()
        creator_cls = PIPELINE_GEVENT_CREATOR_CLASSES.get(gevent.pipeline)
        if creator_cls is None:
            raise NotImplementedError(
                f'Cannot re-create G-event {gevent_id!r} from unknown pipeline '
                f'{gevent.pipeline!r}.'
            )
        return creator_cls(gevent, target, cache_entry)

    async def create(
        self, group: str | None, search: str | None, event_time: float, delay: float = 0
    ) -> dict[str, Any]:
        """Time-translate the original G-event and re-generate it in GraceDB.

        Parameters:
            group: The analysis group which identified the candidate. If None, the
                original group is used.
            search: The type of search of the analysis pipeline. If None, the
                original search type is used.
            event_time: The new intrinsic event time (GPS time).
            delay: Delay to be waited for before sending the creation request to
                GraceDB (in seconds).
        Returns:
            The newly created G-event, as a dictionary.
        """
        await trio.sleep(delay)

        if group is None:
            group = self.gevent.group
        if search is None:
            search = self.gevent.search

        created_event = await self._upload_initial_data(group, search, event_time)
        event_id = created_event['graceid']

        for upload in filter_uploads_unless(self.gevent.uploads, False)[1:]:
            await self._upload_additional_data(event_id, upload)
            await self._write_label(event_id, upload)

        return created_event

    async def _upload_initial_data(
        self, group: str, search: str, event_time: float
    ) -> dict[str, Any]:
        initial_upload = self.gevent.uploads[0]
        initial_filename, _ = split_filename(initial_upload.filename)
        filecontents = self._read_upload(initial_upload)
        delta_time = event_time - self.gevent.gpstime
        shifted_filecontents = self._shift_original_data(filecontents, delta_time)

        try:
            created_event: dict[str, Any] = self.target.create_event(
                pipeline=self.gevent.pipeline,
                group=group,
                search=search,
                offline=self.gevent.offline,
                filename=initial_filename,
                filecontents=shifted_filecontents,
            ).json()
        except HTTPError as exc:
            logger.error(exc.response.text)
            raise

        self._report_creation(created_event)
        return created_event

    def _read_upload(self, upload: Upload) -> bytes:
        return self.cache_entry.read_bytes(upload.filename)

    def _shift_original_data(self, filecontents: bytes, delta_time: float) -> bytes:
        raise NotImplementedError

    def _report_creation(self, created_gevent: dict[str, Any]) -> None:
        id = f'{self.gevent.id} -> {created_gevent["graceid"]}  {self.gevent.pipeline}'
        group = created_gevent['group']
        if group != self.gevent.group:
            group = f'{self.gevent.group} -> {group}'
        search = created_gevent['search']
        if search != self.gevent.search:
            search = f'{self.gevent.search} -> {search}'
        event_time = created_gevent['gpstime']
        logger.info(f'Created {id:31}{group:15}{search:15}{event_time}')

    async def _upload_additional_data(self, event_id: str, upload: Upload) -> None:
        """Upload a file to GraceDB."""
        filename, _ = split_filename(upload.filename)
        await trio.sleep(upload.delay)

        try:
            self.target.write_log(
                event_id,
                filename=filename,
                filecontents=self._read_upload(upload),
                message=upload.message,
                tag_name=upload.tags,
            )
        except HTTPError as exc:
            logger.error(exc.response.text)
            raise
        logger.info(f'Uploaded file for event {event_id}: {upload.filename}')

    async def _write_label(self, event_id: str, upload: Upload) -> None:
        """Annotate an event with a label in GraceDB.

        If it is the first upload and if the file is not a PNG image, the presence of
        the tag 'p_astro' or 'sky_loc' will respectively trigger the
        annotation with the label 'PASTRO_READY' or 'SKYMAP_READY'.
        """
        filename, version = split_filename(upload.filename)
        if version != 0:
            return

        if filename == 'em_bright.json':
            # FIXME disabled for now
            # label = 'EMBRIGHT_READY'
            return
        elif filename == 'p_astro.json':
            label = 'PASTRO_READY'
        elif 'sky_loc' in upload.tags and IS_FITS_REGEX.search(filename):
            label = 'SKYMAP_READY'
        else:
            return

        try:
            self.target.write_label(event_id, label)
        except HTTPError as exc:
            logger.error(exc.response.text)
            raise
        logger.info(f'Written label for event {event_id}: {label}')


class CWBEventCreator(GEventCreator):
    """Creator of G-event for the cWB pipeline."""

    ORIGINAL_DATA_REGEX = re.compile(
        r'(?<=(^start:|^time: |^stop: ))( +(\d+\.\d*))+', re.MULTILINE
    )
    ORIGINAL_DATA_LINE_REGEX = re.compile(r'\d+\.\d*')

    @staticmethod
    def _shift_original_data(raw_content: bytes, delta_time: float) -> bytes:
        def time_repl(match: re.Match[str]) -> str:
            new_time = float(match.group(0)) + delta_time
            return f'{new_time:.4f}'

        def line_repl(match: re.Match[str]) -> str:
            new_line, nsub = CWBEventCreator.ORIGINAL_DATA_LINE_REGEX.subn(
                time_repl, match.group(0)
            )
            if nsub not in {2, 3}:
                raise ValueError(
                    f'Could not extract the event times. Has the format changed?\n'
                    f'{content}'
                )
            return new_line

        content = raw_content.decode()
        new_content, nsub = CWBEventCreator.ORIGINAL_DATA_REGEX.subn(line_repl, content)
        if nsub != 3:
            raise ValueError(
                f'Could not extract the event times. Has the format changed?\n{content}'
            )
        return new_content.encode()


class CBCEventCreator(GEventCreator):
    """Creator of G-event for the CBC pipelines."""

    @staticmethod
    def _shift_original_data(raw_data: bytes, delta_time: float) -> bytes:
        buffer = BytesIO(raw_data)
        xmldoc: Document = ligo.lw.utils.load_fileobj(
            buffer, contenthandler=ContentHandler
        )

        CBCEventCreator._shift_original_document(xmldoc, delta_time)

        buffer = BytesIO()
        ligo.lw.utils.write_fileobj(xmldoc, buffer)
        return buffer.getvalue()

    @staticmethod
    def _shift_original_document(xmldoc: Document, delta_time: float) -> None:
        coinc_inspiral_table = ligo.lw.lsctables.CoincInspiralTable.get_table(xmldoc)
        for row in coinc_inspiral_table:
            row.end += delta_time

        sngl_inspiral_table = ligo.lw.lsctables.SnglInspiralTable.get_table(xmldoc)
        for row in sngl_inspiral_table:
            row.end += delta_time

        arrays = xmldoc.getElementsByTagName('LIGO_LW')
        for array in arrays:
            if not hasattr(array, 'Name'):
                continue
            if array.Name == 'COMPLEX8TimeSeries':
                array.childNodes[0].pcdata += delta_time


def etree_find(root: _ElementTree, path: str) -> _Element:
    """Returns an element in an xml document, or raise an error if not found.

    Parameters:
        root: The xml document.
        path: The specification path to the element.

    Raises:
        ValueError: When the element specified by the path is not specified in document.
    """
    element = root.find(path)
    if element is None:
        raise ValueError(f'No element specified by {path!r} in the xml document.')
    return element


class ExternalEventCreator(GEventCreator):
    """Creator of G-event for External group pipelines."""

    @staticmethod
    def _shift_original_data(raw_data: bytes, delta_time: float) -> bytes:
        buffer = BytesIO(raw_data)
        root = etree.parse(buffer)
        ExternalEventCreator._shift_original_document(root, delta_time)
        return etree.tostring(root, xml_declaration=True, pretty_print=True)

    @staticmethod
    def _shift_original_document(root: _ElementTree, delta_time: float) -> None:
        for tag in [
            './Who/Date',
            './WhereWhen/ObsDataLocation/ObservationLocation/AstroCoords/Time/TimeInstant/ISOTime',  # noqa: E501
        ]:
            element = etree_find(root, tag)
            original_gps_time = Time(element.text, format='isot').gps
            new_gps_time = original_gps_time + delta_time
            new_isot_time = Time(new_gps_time, format='gps', scale='utc').isot
            element.text = str(new_isot_time)


def filter_uploads_unless(
    uploads: list[Upload], include_all_files: bool
) -> list[Upload]:
    """Optionally remove uploads of lesser importance.

    This function optionally filters out the uploads of PNG files or files without one
    of the tags: 'p_astro', 'psd' and 'sky_loc'. The initial upload is always included.

    Parameters:
        uploads: The event uploads.
        include_all_files: If true, the filtering is not performed, otherwise the
            filtering is performed.
    """
    if include_all_files:
        return uploads
    return filter_uploads(uploads)


def filter_uploads(uploads: list[Upload]) -> list[Upload]:
    """Remove uploads of lesser importance.

    This function filters out the uploads of PNG files or files without one of the tags:
    'em_bright', 'p_astro' and 'sky_loc'. The initial upload is always included.

    Parameters:
        uploads: The event uploads.
    """
    required_tags = {'em_bright', 'p_astro', 'sky_loc'}
    new_uploads = [uploads[0]]

    for upload in uploads[1:]:
        if not upload.tags & required_tags:
            continue
        filename, _ = split_filename(upload.filename)
        if filename.endswith('.png'):
            continue
        new_uploads.append(upload)

    return new_uploads


PIPELINE_GEVENT_CREATOR_CLASSES = {
    # CBC pipelines
    'gstlal': CBCEventCreator,
    'MBTAOnline': CBCEventCreator,
    'pycbc': CBCEventCreator,
    'spiir': CBCEventCreator,
    # Burst pipelines
    'CWB': CWBEventCreator,
    # 'MLy': MLyEventCreator,
    # External events pipelines Group=External
    'Fermi': ExternalEventCreator,
    'Swift': ExternalEventCreator,
    'SNEWS': ExternalEventCreator,
    'INTEGRAL': ExternalEventCreator,
    'AGILE': ExternalEventCreator,
}



"""This module provides the functionality to access cached event cached data files."""

from __future__ import annotations

import json
import logging
import os
from dataclasses import dataclass
from pathlib import Path
from typing import Any

from apischema import deserialize

from mock_event_generator.gracedbs import GraceDBWithContext

from .fetchers import GEventFetcher, SEventFetcher
from .models import GEventDescription, SEventDescription

__all__ = ['EventFileCache', 'GEventCacheEntry', 'SEventCacheEntry']

logger = logging.getLogger(__name__)

try:
    DEFAULT_CACHE_PATH = Path(os.environ['XDG_CACHE_HOME']) / 'mock-event-generator'
except KeyError:
    DEFAULT_CACHE_PATH = Path.home() / '.cache' / 'mock-event-generator'


@dataclass(frozen=True)
class EventFileCache:
    """Class to cache and access G-event or S-event data files.

    Attributes:
        source: The GraceDB instance name from which events are downloaded, such as
            `production` or `playground`.
        disabled: If true, bypass the cache and always download the event data files.
        cache_path: The top-level path of the cache.
    """

    source: GraceDBWithContext
    disabled: bool
    cache_path: Path = DEFAULT_CACHE_PATH

    def __post_init__(self) -> None:
        """Ensure that the cache path exists."""
        self.cache_path.mkdir(exist_ok=True, parents=True)

    def get_sevent_cache_entry(self, sevent_id: str) -> SEventCacheEntry:
        """Factory to cache and access S-event data files.

        If the data is not available in the cache, it is immediately downloaded.

        Parameters:
            sevent_id: Identifier of the S-event to be downloaded.
        """
        return SEventCacheEntry.from_id(
            sevent_id, self.source, self.disabled, self.cache_path
        )

    def get_gevent_cache_entry(self, gevent_id: str) -> GEventCacheEntry:
        """Factory to cache and access G-event data files.

        If the data is not available in the cache, it is immediately downloaded.

        Attributes:
            gevent_id: Identifier of the G-event to be downloaded.
        """
        try:
            path = next(self.cache_path.glob(f'*/{gevent_id}')).parent
        except StopIteration:
            path = self.cache_path

        return GEventCacheEntry.from_id(gevent_id, self.source, self.disabled, path)


@dataclass(frozen=True)
class CacheEntry:
    """Class to access cached data files.

    Attributes:
        path: The path to the cache entry.
    """

    path: Path

    def read_bytes(self, filename: str) -> bytes:
        return (self.path / filename).read_bytes()

    def read_json(self, filename: str) -> Any:
        return json.loads(self.read_bytes(filename))


class SEventCacheEntry(CacheEntry):
    """Class to access cached S-event data files.

    Attributes:
        path: The path to the cache entry.
    """

    @classmethod
    def from_id(
        cls,
        sevent_id: str,
        source: GraceDBWithContext,
        disabled: bool,
        cache_path: Path,
    ) -> SEventCacheEntry:
        """Factory to cache and access S-event data files.

        If the data is not available in the cache, it is downloaded.

        Parameters:
            sevent_id: Identifier of the S-event to be downloaded.
            source: The GraceDB client from which events are downloaded.
            disabled: If true, bypass the cache and always download the data files.
            cache_path: The top-level path of the cache.
        """
        sevent_path = cache_path / sevent_id
        if disabled or not sevent_path.is_dir():
            logger.info(
                f'Downloading {sevent_id} from GraceDB ('
                f'{source.meg_alias or source.meg_url})...'
            )
            SEventFetcher.from_id(sevent_id, source, sevent_path.parent).fetch()
        return SEventCacheEntry(sevent_path)

    def __truediv__(self, gevent_id: str) -> GEventCacheEntry:
        """Get a G-Event cache entry from an S-Event cache entry.

        Parameters:
            gevent_id: The G-event identifier.

        Example:
            >>> cache_entry = SEventCacheEntry(DEFAULT_CACHE_PATH / 'S00001')
            >>> cache_entry / 'G00001'
            GEventCacheEntry(
                path=PosixPath('/home/user/.cache/mock_event_generator/S00001/G00001')
            )
        """
        if not isinstance(gevent_id, str):
            return NotImplemented
        return GEventCacheEntry(self.path / gevent_id)

    def get_description(self) -> SEventDescription:
        """Return the serialized description of the S-event."""
        serialized_description = self.read_json('description.json')
        return deserialize(SEventDescription, serialized_description)


class GEventCacheEntry(CacheEntry):
    """Class to access cached G-event data files.

    Attributes:
        path: The path to the cache entry.
    """

    @classmethod
    def from_id(
        cls,
        gevent_id: str,
        source: GraceDBWithContext,
        disabled: bool,
        cache_path: Path,
    ) -> GEventCacheEntry:
        """Factory to cache and access G-event data files.

        If the data is not available in the cache, it is downloaded.

        Attributes:
            gevent_id: Identifier of the G-event to be downloaded.
            source: The GraceDB client from which events are downloaded.
            disabled: If true, bypass the cache and always download the data files.
            cache_path: The top-level or associated S-event cache path.
        """
        gevent_path = cache_path / gevent_id
        if disabled or not gevent_path.is_dir():
            logger.info(
                f'Downloading {gevent_id} from GraceDB ('
                f'{source.meg_alias or source.meg_url})...'
            )
            GEventFetcher.from_id(gevent_id, source, gevent_path.parent).fetch()
        return GEventCacheEntry(gevent_path)

    def get_description(self) -> GEventDescription:
        """Return the serialized description of the G-event."""
        serialized_description = self.read_json('description.json')
        return deserialize(GEventDescription, serialized_description)



"""The command line interface for the package `mock_event_generator`."""
import shutil
import sys
from datetime import datetime
from pathlib import Path
from typing import Optional

import certifi
import trio
from astropy.time import Time
from cryptography import x509
from cryptography.hazmat.backends import default_backend
from typer import Argument, Exit, Option, Typer, echo

from mock_event_generator.exceptions import MEGInvalidGraceDBAliasOrURLError

from .cache import DEFAULT_CACHE_PATH, EventFileCache, GEventCacheEntry
from .creators import GEventCreator, SEventCreator
from .gracedbs import GraceDBAlias, GraceDBWithContext
from .utils import is_any_event, is_gevent, is_superevent, tree

meg = Typer(help='Mock Event Generator.')
GRACEDB_ALIASES = ', '.join(alias.value for alias in GraceDBAlias)


@meg.command()
def create(
    events: list[str] = Argument(..., help='G-events or S-events to be generated.'),
    target: str = Option(
        ...,
        help=f'GraceDB instance ({GRACEDB_ALIASES} or <URL>) to which the time-'
        'translated events are sent.',
    ),
    username: Optional[str] = Option(
        None, help='Username for basic authentication on the target GraceDB server.'
    ),
    password: Optional[str] = Option(
        None, help='Password for basic authentication on the target GraceDB server.'
    ),
    source: str = Option(
        GraceDBAlias.PRODUCTION,
        help=f'GraceDB instance ({GRACEDB_ALIASES} or <URL>) from which the original '
        'events are downloaded.',
    ),
    group: Optional[str] = Option(
        None,
        help='Change the analysis group which identified the candidate.',
    ),
    search: Optional[str] = Option(
        None,
        help='Change the type of search of the analysis pipeline. By default, '
        "the event search is changed to 'MDC'.",
    ),
    original_search: bool = Option(
        False, help='Use the original event search type, instead of MDC.'
    ),
    cache_path: Path = Option(
        DEFAULT_CACHE_PATH, help="Directory where the event' data files are downloaded."
    ),
    refresh_cache: bool = Option(
        False, help="If set, ignore the event's potential cache entry."
    ),
    max_delay: Optional[float] = Option(
        None,
        help='Shrink the interval between the first event creation and the last upload '
        '(in seconds). By setting zero, all uploads are sent at once.',
    ),
) -> None:
    """Create G-events and send them to GraceDB."""
    if search is None and not original_search:
        search = 'MDC'

    _check_event_ids(events)
    try:
        source_client = GraceDBWithContext.meg_from_alias_or_url(source)
        target_client = GraceDBWithContext.meg_from_alias_or_url(
            target, username=username, password=password
        )
    except MEGInvalidGraceDBAliasOrURLError as exc:
        echo(exc, err=True, color=True)
        raise Exit(1)

    cache = EventFileCache(source_client, refresh_cache, cache_path)
    now = Time.now().gps

    async def create_all() -> None:
        async with trio.open_nursery() as nursery:
            for event in events:
                if is_superevent(event):
                    nursery.start_soon(
                        SEventCreator.from_id(event, target_client, cache).create,
                        group,
                        search,
                        max_delay,
                    )
                else:
                    nursery.start_soon(
                        GEventCreator.from_id(event, target_client, cache).create,
                        group,
                        search,
                        now,
                        0,
                    )

    trio.run(create_all)


@meg.command()
def fetch(
    events: list[str] = Argument(..., help='G-events or S-events to be generated.'),
    source: str = Option(
        GraceDBAlias.PRODUCTION,
        help=f'GraceDB instance ({GRACEDB_ALIASES} or <URL>) from which the original '
        'events are downloaded.',
    ),
    cache_path: Path = Option(
        DEFAULT_CACHE_PATH, help="Directory where the event' data files are downloaded."
    ),
    refresh_cache: bool = Option(
        False, help="If set, ignore the event's potential cache entry."
    ),
) -> None:
    """Fetch G-events and store them in the cache."""
    _check_event_ids(events)
    try:
        source_client = GraceDBWithContext.meg_from_alias_or_url(source)
    except MEGInvalidGraceDBAliasOrURLError as exc:
        echo(exc, err=True, color=True)
        raise Exit(1)

    cache = EventFileCache(source_client, refresh_cache, cache_path)
    for event in events:
        if is_superevent(event):
            cache.get_sevent_cache_entry(event)
        else:
            cache.get_gevent_cache_entry(event)


cache = Typer()
meg.add_typer(cache, name='cache', help='Event cache utilities')


@cache.command()
def clean(
    cache_path: Path = Option(
        DEFAULT_CACHE_PATH, help="Directory where the event' data files are downloaded."
    ),
) -> None:
    """Remove the content of the cache."""
    if not cache_path.is_dir():
        echo(f'Cache path does not exist: {cache_path}', err=True)
        sys.exit(1)
    print(f'Cleaning cache: {cache_path}')
    for path in cache_path.iterdir():
        if path.is_dir() and is_any_event(path.name):
            print(f'Removing {path}')
            shutil.rmtree(path)


@cache.command('list')
def list_(
    include_files: bool = Option(False, help='If set, also display the data files.'),
    cache_path: Path = Option(
        DEFAULT_CACHE_PATH, help="Directory where the event' data files are downloaded."
    ),
) -> None:
    """List the content of the cache."""
    if not cache_path.is_dir():
        echo(f'Cache path does not exist: {cache_path}', err=True)
        sys.exit(1)
    if include_files:

        def criterion(path: Path) -> bool:
            return True

    else:

        def criterion(path: Path) -> bool:
            return path.is_dir()

    def sort_key(path: Path) -> float:
        """Sort key according to modification time (older first)."""
        return 0 if path.name == 'description.json' else path.stat().st_mtime

    print(f'Cache: {cache_path}')
    for path, line in tree(cache_path, criterion, key=sort_key):
        is_dir = path.is_dir()
        if is_dir and is_gevent(path.name):
            entry = GEventCacheEntry(path)
            description = entry.get_description()
            line = line.ljust(20) + ''.join(
                str(_).ljust(15)
                for _ in [
                    description.pipeline,
                    description.group,
                    description.search,
                    description.gpstime,
                ]
            )
        print(line)


@meg.command()
def ca_certificate(path: Path = Argument(..., help='The CA certificate path.')) -> None:
    """Add a Certification Authority certificate.

    The certificate is added to the CA bundle used by the requests library.
    """
    content = path.read_bytes()
    cert = x509.load_pem_x509_certificate(content, default_backend())
    if cert.not_valid_after < datetime.now():
        echo(f'The CA certificate {path.name} has expired.', err=True, color=True)
        raise Exit(1)

    original_content = Path(certifi.where())
    if content in original_content.read_bytes():
        echo(f'The CA certificate {path.name} has already been added.')
        raise Exit()

    with Path(certifi.where()).open('ba') as f:
        f.write(b'\n\n')
        f.write(content)
        f.write(b'\n')


def _check_event_ids(events: list[str]) -> None:
    """Abort if any of the input event identifier is invalid."""
    invalid_event_ids = [repr(_) for _ in events if not is_any_event(_)]
    if not invalid_event_ids:
        return

    echo(
        f'Invalid event identifier(s): {", ".join(invalid_event_ids)}.',
        err=True,
        color=True,
    )
    raise Exit(1)



import sys
from pathlib import Path
from typing import Iterator

import pytest
from pytest_mock import MockerFixture

from mock_event_generator.gracedbs import GraceDBAlias, GraceDBWithContext

SOURCE = GraceDBWithContext.meg_from_alias_or_url(GraceDBAlias.PLAYGROUND)

# include the tests directory in the path so that we can import tests from it
sys.path.insert(0, str(Path(__file__).parents[1]))


@pytest.fixture
def mocked_gracedb(mocker: MockerFixture) -> Iterator[None]:
    mocker.patch('tests.test_fetchers_gevent.SOURCE', SOURCE)
    mocker.patch('tests.test_fetchers_sevent.SOURCE', SOURCE)
    yield
